/***************************************************************************************/
/* Demo for driving PIC16Fxx as I2C-Master according to Application Note AN734         */
/*                                                                                     */
/* Master                                                                              */
/*                                                                                     */
/* Description: This example is indended to demonstrate several capabilities of I2C bus*/
/*              - master/slave communication                                           */
/*              - timeout in case of communiation failure                              */
/*              - capability of sending unsigned char variables to 32 different        */
/*                internal UC registers                                                */
/*              - variable data lenght                                                 */
/*                                                                                     */
/* Mikrocontroller:       PIC16F887                                                    */
/* Clock:                 internal clock (8.0 MHz)                                     */
/*                                                                                     */
/* Compiler:              mikroC V6.6.2                                                */
/*                                                                                     */
/* Developer:            Mitroi Nicolae Alexandru                                      */
/* Stert of software development: 01. August 2016                                      */
/* Last updated:                  07. August 2016                                      */
/***************************************************************************************/


/******************************************************************************
#definitions section
******************************************************************************/
 #define _I2C_MAX_REGISTERS_COUNT  32 //Holds the max number of internal registers where master can write data
 #define _SLAVE_MCU_REGISTER_MASK  31 //Holds the mask for the slave registers; This should be set according to _I2C_MAX_REGISTERS_COUNT register
 #define _MSG_COUNT_SHIFT_NR_POS    7 //Holds the nr of shift position to write message number
 #define _HEADER_BIT               64 //Holds the header bit
 #define _HEADER_READ_REQUESTED    32 //Holds the header bit
 #define _MESSAGE_MASK             15 //Holds the message mask
 #define _ACKANOWLEDGE              1 //Holds the acknowledge bit (it will be send to slave on read)
 #define _NO_ACKANOWLEDGE           0 //Holds the no acknowledge bit (it will be send to slave on read)
 #define _I2C_TIMEOUT             200 //Holds the read/write timeout


/******************************************************************************
Function: I2C_WriteEx
Return:
Arguments: dat        - 8-bit data to be sent on bus; data can be either
                        address/data byte
           timeout_us - interval of time in us after which the function exits
                        in case of failure
Description: Send 8-bit data on I2C bus
******************************************************************************/
int I2C_WriteEx(unsigned char dat, unsigned long timeout_us)
{
     int result = 0; // Set result to OK
     Delay_us(100);
     PIR1 = PIR1&0b11110111;  //Reset SSPIF bit
     SSPBUF = dat;  //Write data to output buffer
     while (R_NOT_W==0 && timeout_us > 0) //while (transmit in progress and timeout not reached)
     {
           timeout_us-=1;
           if (R_NOT_W==0) //if (transmit in progress)
           {
              Delay_us(1);
           }
     }
     if (timeout_us==0)
     {
        PORTA=PORTA|0b00000001;
        SSPCON2=SSPCON2|0b00000100; //PEN bit = 1;
        result = -1; // Set result to -1 (TIMEOUT)
        goto _exitPoint;
     }
     while (SSPIF==0 && timeout_us > 0) //while(no interrupt occured and timeout not reached)
     {
           timeout_us-=1;
           if (SSPIF==0) //if(no interrupt occured)
           {
              Delay_us(1);
           }
     }
     if (timeout_us==0)
     {
        PORTA=PORTA|0b00000010;
        SSPCON2=SSPCON2|0b00000100; //PEN bit = 1;
        result = -1; // Set result to -1 (TIMEOUT)
        goto _exitPoint;
     }
     if (ACKSTAT==0) //if(acknowledge was received)
     {
      asm nop;
     }
     {
         SSPCON2=SSPCON2|0b00000100; //PEN bit = 1;
         result = -2; // Set result to -2 (Not Acknowledged)
     }
_exitPoint:
     return result;
}


/******************************************************************************
Function: I2C_ReadEx
Return:    8-bit data read from I2C bus
Arguments: acknowledge - if 0 then no acknowledge is send else acknowledge
                         is send
           timeout_us  - interval of time in us after
                         which the function exits in case of failure
Description: read 8-bit data from I2C bus
******************************************************************************/
int I2C_ReadEx(char acknowledge, unsigned long timeout_us)
{
    int result = -1; // Set result to -1 (TIMEOUT)
    Delay_us(100);
    PIR1 = PIR1&0b11110111;  //Reset SSPIF bit
    SSPCON2 = SSPCON2|0b00001000; //Enable receive mode
    while(timeout_us>0 && !(PIR1&0b00001000)) //while(timeout not reached and no interrupt occured)
    {
        timeout_us-=1;
        if (!(PIR1&0b00001000)) //if(no interrupt occured)
        {
             Delay_us(1);
        }
    }
    if (timeout_us==0)
    {
        PORTA=PORTA|0b00000100;
        SSPCON2=SSPCON2|0b00000100; //PEN bit = 1;
    }
    else
    {
        result = SSPBUF;
    }
    if (acknowledge==0)
    {
        SSPCON2 = SSPCON2|0b00100000;  // Set SSPCON2.ACKDT (Not Acknowledge)
    }
    else
    {
        SSPCON2 = SSPCON2&0b11011111; // Reset SSPCON2.ACKDT (Acknowledge)
    }
    PIR1 = PIR1&0b11110111;  //Reset SSPIF bit
    SSPCON2 = SSPCON2|0b00010000;  //Send ack/no ack back to slave
    while(timeout_us>0 && !(PIR1&0b00001000))  //while(timeout not reached and no interrupt occured)
    {
        timeout_us-=1;
        if (!(PIR1&0b00001000))  //if(no interrupt occured)
        {
             Delay_us(1);
        }
    }
    if (timeout_us==0)
    {
        PORTA=PORTA|0b00001000;
        SSPCON2=SSPCON2|0b00000100; //PEN bit = 1;
        result = -1; // Set result to -1 (TIMEOUT)
    }
    return result;
}


/******************************************************************************
Function: I2C_MasterWriteEx
Return:
Arguments: chipAddress    - represent the 7 bits slave address
           slaveRegister  - 5 -bit register (internal variables) from slave MCU
           input          - data to be transmitted
Description: Send unsigned char data on I2C bus
Warning!!! Please note that there is a difference between a phisical read/write send to chip
(e.g. chipAddress|1 means a phisical read is requested to slave) and logical read/write
which is set on header
******************************************************************************/
int I2C_MasterWriteEx(unsigned char chipAddress, unsigned char slaveRegister, unsigned char input)
{
     int result = 0; // Set result to OK
     unsigned char cdata;
     unsigned char header = 0;
     char count;
     
     I2C1_Start();
     chipAddress = chipAddress<<1;
     I2C_WriteEx(chipAddress,_I2C_TIMEOUT);
     header = _HEADER_BIT;
     header = header | 0b10000000; // Inform the slave that two more requests will be received; If 6th bit is 0 then inform the slave that a logical write will be performed
     header = header | (slaveRegister & _SLAVE_MCU_REGISTER_MASK);
     result = I2C_WriteEx(header,_I2C_TIMEOUT);
     I2C1_Stop();
     
     I2C1_Start();
     I2C_WriteEx(chipAddress|1,_I2C_TIMEOUT); // A phisical read is requested
     result = I2C_ReadEx(_NO_ACKANOWLEDGE,_I2C_TIMEOUT);
     I2C1_Stop();
     
     if (result>=0 && result == header) //The header was successfully transmitted to slave
     {
       for(count = 0; count<2; count++)
       {
          if (count==0)
          {
             cdata = input&_MESSAGE_MASK;    // If 8th bit is 0 then the first 4 bites are writted
          }
          else
          {
              cdata = (input>>4)&_MESSAGE_MASK;
              cdata = cdata | 0b10000000;    // If 8th bit is 1 then the second 4 bites are writted
          }
          I2C1_Start();
          result = I2C_WriteEx(chipAddress,_I2C_TIMEOUT);  // A phisical write is requested
          result = I2C_WriteEx(cdata,_I2C_TIMEOUT);
          I2C1_Stop();
          
          I2C1_Start();
          I2C_WriteEx(chipAddress|1,_I2C_TIMEOUT);  // A phisical read is requested
          result = I2C_ReadEx(_NO_ACKANOWLEDGE,_I2C_TIMEOUT);
          I2C1_Stop();
     
          if (result<0 || result != cdata)
          {
             break;
          }
       }
     }
     return result;
}


/******************************************************************************
Function: I2C_MasterReadEx
Return: int
Arguments: chipAddress    - represent the 7 bits slave address
           slaveRegister  - 5 -bit register (internal variables) from slave MCU
Description: Receive unsigned char from I2C bus
Warning!!! Please note that there is a difference between a phisical read/write send to chip
(e.g. chipAddress|1 means a phisical read is requested to slave) and logical read/write
which is set on header
******************************************************************************/
int I2C_MasterReadEx(unsigned char chipAddress, unsigned char slaveRegister)
{
     unsigned char result = 0;
     char counter = 0;
     char header = 0;
     int tempResult = -1;
     chipAddress = chipAddress<<1;
     for (counter=0; counter<2; counter++)
     {
         header = counter<<_MSG_COUNT_SHIFT_NR_POS; // Write the message number to be readed from slave (a byte has 2 pair of 4 bites)
         header = header | _HEADER_BIT;
         header = header | _HEADER_READ_REQUESTED;  // Logical read was requested from slave
         header = header | (slaveRegister & _SLAVE_MCU_REGISTER_MASK);
         I2C1_Start();
         I2C_WriteEx(chipAddress,_I2C_TIMEOUT); // Phisical write was requested from slave
         I2C_WriteEx(header,_I2C_TIMEOUT);
         I2C1_Stop();
         I2C1_Start();
         I2C_WriteEx(chipAddress|1,_I2C_TIMEOUT); //Phisical read was requested from slave
         tempResult = I2C_ReadEx(_NO_ACKANOWLEDGE,_I2C_TIMEOUT);
         I2C1_Stop();
         if (tempResult<0) // An error was encountered in transmission
         {
            result = tempResult;
            return;
         }
         else
         {
             tempResult = tempResult & _MESSAGE_MASK;
             tempResult = tempResult << (counter * 4);
             result = result | tempResult;
         }
     }
     return result;
}