/***************************************************************************************/
/* Demo for driving PIC16Fxx as I2C-Slave according to Application Note AN734          */
/*                                                                                     */
/* Slave                                                                               */
/*                                                                                     */
/* Description: This example is indended to demonstrate several capabilities of I2C bus*/
/*              - master/slave communication                                           */
/*              - timeout in case of communiation failure                              */
/*              - capability of sending unsigned char variables to 32 different        */
/*                internal UC registers                                                */
/*              - variable data lenght                                                 */
/*                                                                                     */
/* Mikrocontroller:       PIC16F887                                                    */
/* Clock:                 internal clock (8.0 MHz)                                     */
/*                                                                                     */
/* Compiler:              mikroC V6.6.2                                                */
/*                                                                                     */
/* Developer:            Mitroi Nicolae Alexandru                                      */
/* Stert of software development: 01. August 2016                                      */
/* Last updated:                  05. August 2016                                      */
/***************************************************************************************/


/******************************************************************************
#definitions section
******************************************************************************/
 #define _I2C_CHIP_ADDRESS         63 //Holds the address received from I2C
 #define _I2C_MAX_REGISTERS_COUNT  32 //Holds the max number of internal registers where master can write data
 #define _SLAVE_MCU_REGISTER_MASK  31 //Holds the mask for the slave registers; This should be set according to _I2C_MAX_REGISTERS_COUNT register
 #define _MSG_COUNT_SHIFT_NR_POS    7 //Holds the nr of shift position to obtain message number
 #define _HEADER_BIT               64 //Holds the header bit
 #define _HEADER_READ_REQUESTED    32 //Holds the header bit
 #define _MESSAGE_MASK             15 //Holds the message mask


/******************************************************************************
Global Variables
******************************************************************************/
extern  char cI2CAddress;                                            //Holds the address received from I2C
extern  char cI2CData;                                               //Holds the last byte received from I2C
extern  unsigned char ucI2CRegisters[_I2C_MAX_REGISTERS_COUNT];      //Holds the register where the data will be stored
extern  unsigned char ucI2CRegistersStatus[_I2C_MAX_REGISTERS_COUNT];//Holds the status of the I2C communication per register
extern  char cI2CCurrentRegisterNumber;                              //Holds the current register number from where the read/write is performed
extern  char cI2CTotalMsgCount;                                      //Holds the total number of messages that should be send/received
extern  char cI2CCurrentMsgCount;                                    //Holds the last message count received
extern  char ucI2CCurrentStatus;                                     //Holds the current I2C status
extern  char cI2CR_W_Operation;                                      //Holds the current operation requested by master


/******************************************************************************
Global Constants
******************************************************************************/
extern const unsigned char _I2C_WRITE = 0;
extern const unsigned char _I2C_READ = 1;

extern const unsigned char _I2C_STATUS_IDLE = 0;
extern const unsigned char _I2C_STATUS_STARTED = 1;
extern const unsigned char _I2C_STATUS_INPROGRESS = 2;
extern const unsigned char _I2C_STATUS_ERROR = 3;
extern const unsigned char _I2C_STATUS_END = 4;

extern const unsigned char _I2C_ADDRESS = 0;
extern const unsigned char _I2C_DATA = 1;


/***************************************************************************************/
/* Function: PrepareResponseData                                                       */
/*                                                                                     */
/* Description:                                                                        */
/*              Prepare data to be write to master                                     */
/*                                                                                     */
/* Return: unsigned char - represents the response that will be send to master         */
/*                                                                                     */
/* Arguments: input (unsigned char) - represents the header message that was           */
/*                                    received from I2C                                */
/***************************************************************************************/
unsigned char PrepareResponseData(unsigned char input)
{
     unsigned char result = 0;
     char receivedMessageCount;
     unsigned char tempVariable;
     receivedMessageCount =  input >> _MSG_COUNT_SHIFT_NR_POS;
     if (cI2CCurrentRegisterNumber>=0 && cI2CCurrentRegisterNumber<_I2C_MAX_REGISTERS_COUNT)
     {
        tempVariable = ucI2CRegisters[cI2CCurrentRegisterNumber];
        tempVariable = tempVariable>>(receivedMessageCount*4);
        tempVariable = tempVariable&_MESSAGE_MASK;
        tempVariable = tempVariable | (receivedMessageCount<<_MSG_COUNT_SHIFT_NR_POS);
        result = tempVariable;
     }
     return result;
}


/***************************************************************************************/
/* Function: ParseI2CHeader                                                            */
/*                                                                                     */
/* Description:                                                                        */
/*              Parse I2C header to obtain the message count, register number and      */
/*              the read/write operation                                               */
/*                                                                                     */
/* Return: char - if -1 then input was not header, if >0 then input was header         */
/*                and was parsed                                                       */
/*                                                                                     */
/* Arguments: input (unsigned char) - represents the header message that was           */
/*                                    received from I2C                                */
/***************************************************************************************/
char ParseI2CHeader(unsigned char input)
{
     char result = -1; // Set result as invalid header
     if (input & _HEADER_BIT)
     {
        cI2CCurrentRegisterNumber = input&_SLAVE_MCU_REGISTER_MASK; // Set current register where the operation will be performed
        cI2CR_W_Operation = input&_HEADER_READ_REQUESTED; // Set logical read/write
        if (input&_HEADER_READ_REQUESTED) //If logical read was requested
        {
           cI2CData = PrepareResponseData(input);
           result = 1;
        }
        else // A logical write was requested
        {
            if (cI2CCurrentRegisterNumber>=0 && cI2CCurrentRegisterNumber<_I2C_MAX_REGISTERS_COUNT)
            {
               ucI2CCurrentStatus = _I2C_STATUS_STARTED;
               ucI2CRegistersStatus[cI2CCurrentRegisterNumber] = _I2C_STATUS_STARTED;
               ucI2CRegisters[cI2CCurrentRegisterNumber] = 0;
               cI2CTotalMsgCount = input>>_MSG_COUNT_SHIFT_NR_POS;
               result = 1;
            }
        }
     }
     return result;
}


/***************************************************************************************/
/* Function: ParseI2CData                                                              */
/*                                                                                     */
/* Description:                                                                        */
/*              Parse I2C data to obtain part of the register                          */
/*                                                                                     */
/* Return: char - if -1 then input was not data, if >0 then input was valid data       */
/*                and was parsed                                                       */
/*                                                                                     */
/* Arguments: input (unsigned char) - represents the data message that was             */
/*                                    received from I2C                                */
/***************************************************************************************/
char ParseI2CData(unsigned char input)
{
     char result = -1; // Set result as invalid data
     unsigned char mask;
     char receivedMessageCount = -1;
     if (input & _HEADER_BIT)
     {
        ucI2CCurrentStatus = _I2C_STATUS_ERROR;
     }
     else
     {
        if (cI2CCurrentRegisterNumber>=0 && cI2CCurrentRegisterNumber<_I2C_MAX_REGISTERS_COUNT)
        {
           mask = input & _MESSAGE_MASK;
           receivedMessageCount =  input >> _MSG_COUNT_SHIFT_NR_POS;
           cI2CCurrentMsgCount+=1;
           if (receivedMessageCount!=cI2CCurrentMsgCount || cI2CCurrentMsgCount > cI2CTotalMsgCount)
           {
              ucI2CCurrentStatus = _I2C_STATUS_ERROR;
              ucI2CRegistersStatus[cI2CCurrentRegisterNumber] = _I2C_STATUS_ERROR;
              ucI2CRegisters[cI2CCurrentRegisterNumber] = 0;
              cI2CCurrentRegisterNumber = -1;
              cI2CTotalMsgCount = -1;
              cI2CCurrentMsgCount = -1;
           }
           else
           {
               mask = mask<<(4*cI2CCurrentMsgCount);
               ucI2CRegisters[cI2CCurrentRegisterNumber] = ucI2CRegisters[cI2CCurrentRegisterNumber] | mask;
               if (cI2CCurrentMsgCount < cI2CTotalMsgCount)
               {
                  ucI2CCurrentStatus = _I2C_STATUS_INPROGRESS;
                  ucI2CRegistersStatus[cI2CCurrentRegisterNumber] = _I2C_STATUS_INPROGRESS;
                  result = 1;
               }
               else if (cI2CCurrentMsgCount == cI2CTotalMsgCount)
               {
                  ucI2CCurrentStatus = _I2C_STATUS_END;
                  ucI2CRegistersStatus[cI2CCurrentRegisterNumber] = _I2C_STATUS_END;
                  cI2CCurrentRegisterNumber = -1;
                  cI2CTotalMsgCount = -1;
                  cI2CCurrentMsgCount = -1;
                  result = 1;
               }
           }
        }
     }
     return result;
}


/***************************************************************************************/
/* Function: ParseI2CInput                                                             */
/*                                                                                     */
/* Description:                                                                        */
/*              Parse I2C input and write it to the corresponding register             */
/*                                                                                     */
/* Return: none                                                                        */
/*                                                                                     */
/* Arguments: input (unsigned char) - represents the input  message that was           */
/*                                    received from I2C                                */
/***************************************************************************************/
void ParseI2CInput(unsigned char input)
{
    if (ucI2CCurrentStatus == _I2C_STATUS_IDLE || ucI2CCurrentStatus == _I2C_STATUS_ERROR || ucI2CCurrentStatus == _I2C_STATUS_END)
    {
         ParseI2CHeader(input);
    }
    else if (ucI2CCurrentStatus == _I2C_STATUS_STARTED || ucI2CCurrentStatus == _I2C_STATUS_INPROGRESS)
    {
         if ((cI2CCurrentRegisterNumber>=0 && cI2CCurrentRegisterNumber<_I2C_MAX_REGISTERS_COUNT)
             &&(ucI2CRegistersStatus[cI2CCurrentRegisterNumber] == _I2C_STATUS_STARTED || ucI2CRegistersStatus[cI2CCurrentRegisterNumber] == _I2C_STATUS_INPROGRESS))
             {
                 ParseI2CData(input);
             }
             else
             {
                 ucI2CCurrentStatus = _I2C_STATUS_ERROR;
             }
    }
    if (ucI2CCurrentStatus == _I2C_STATUS_END)
    {
       ucI2CCurrentStatus = _I2C_STATUS_IDLE;
    }
}


/***************************************************************************************/
/* Function: ManageI2CBus                                                              */
/*                                                                                     */
/* Description:                                                                        */
/*              Manage I2C bus via MCU interrupts when a transmission occurs           */
/*                                                                                     */
/* Return: none                                                                        */
/*                                                                                     */
/* Arguments: none                                                                     */
/***************************************************************************************/
void ManageI2CBus()
{
    if (SSPSTAT.R_W == _I2C_WRITE) //The W/R bit (0) indicate a write
    {
        // This indicates that the master wants to write some data
        // (address or data) to the slave
        if (SSPSTAT.D_A == _I2C_ADDRESS) //The last received byte was an address
        {
            // State 1: I2C write operation, last byte was an address byte
            // SSPSTAT bits: S=1, D/A=0, R/W=0, BF=1
            cI2CAddress = SSPBUF;
            SSPSTAT.BF = 0;                    // Reset flag Buffer-Full
        }
        else //The last received byte was data
        {
            // State 2: I2C write operation, last byte was a data byte
            // SSPSTAT bits: S=1, D/A=1, R/W=1, BF=1
            cI2CData = SSPBUF;                    // Read data from SSPBUF
            //SSPSTAT.BF = 0;                    // Reset flag Buffer-Full
            ParseI2CInput(cI2CData);
        }
    }
    else //The W/R bit (1) indicate a read
    {
        // This indicates that the master has requested some data
        if (SSPSTAT.D_A == _I2C_ADDRESS) //The last received byte was an address
        {
            // State 3: I2C read operation, last byte was an address byte
            // SSPSTAT bits: S=1, D/A=0, R/W=1, BF=0
            SSPBUF = cI2CData;                    // Write data from cI2CData to SSPBUF
        }
        else  //The last received byte was data
        {
            // State 4: I2C read operation, last byte was a data byte
            // SSPSTAT bits: S=1, D/A=1, R/W=0, BF=0
            SSPBUF = cI2CAddress;              // Respond with this chip address
        }
    }
    SSPCON.CKP = 1;                       // Allow clock stretching
    PIR1.SSPIF = 0;                       // Reset SSP-Interrupt-Flag
}


/***************************************************************************************/
/* Function: InitI2CSlave                                                              */
/*                                                                                   */
/* Description:                                                                        */
/*   Initialize the MCU:                                                               */
/*      - Configure I2C Hardware module for slave                                      */
/*      - Write I2C slave address from _I2C_CHIP_ADDRESS constant                      */
/*      - Configure interrupts accordingly                                             */
/*                                                                                     */
/* Return: nothing                                                                     */
/*                                                                                     */
/* Arguments: none                                                                     */
/***************************************************************************************/
void InitI2CSlave(void)
{
     char counter;
     for (counter = 0; counter<_I2C_MAX_REGISTERS_COUNT; counter++)
     {
         ucI2CRegisters[counter] = 0;
     }
     cI2CCurrentRegisterNumber = -1;
     cI2CTotalMsgCount = -1;
     cI2CCurrentMsgCount = -1;
     ucI2CCurrentStatus = _I2C_STATUS_IDLE;
     
    // (internal) oscilator configuration
    OSCCON = 0b01110001;             // Sync Oscillator Control Register
/*             +-------------------- Bit 7 Reserved
                +++----------------- Bit 6-4 (IRCF2:IRCF0): Internal Oscillator Frequency
                   ||||                       Select bits
                   ||||                 000 : 31 kHz (LFINTOSC)
                   ||||                 001 : 125 kHz
                   ||||                 010 : 250 kHz
                   ||||                 011 : 500 kHz
                   ||||                 100 : 1 MHz
                   ||||                 101 : 2 MHz
                   ||||                 110 : 4 MHz (default)
                   ||||              -> 111 : 8 MHz
                   +---------------- Bit 3 (OSTS): Oscillator Start-up Time-out Status
                    |||                    (Read Only)
                    |||                   0 : Device is running from the clock defined
                    |||                       by FOSC2:FOSC0 of the CONFIG1 register
                    |||                   1 : Device is running from the internal
                    |||                       oscillator  (HFINTOSC or LFINTOSC)
                    +--------------- Bit 2 (HTS): HFINTOSC Status bir
                     ||                    (High Frequency - 125kHz to 8MHz)
                     ||                    (Read Only)
                     ||                   0 : HFINTOSC is not stable
                     ||                   1 : HFINTOSC is stable
                     +-------------- Bit 1 (LTS): LFINTOSC Status bit
                      |                    (Low Frequency - 31kHz)
                      |                    (Read Only)
                      |                   0 : LFINTOSC is not stable
                      |                   1 : LFINTOSC is stable
                      +------------- Bit 0 (SCS): System Clock Select bit
                                          0 : Clock source defined by FOSC2:FOSC0
                                              of the CONFIG1 register
                                     ->   1 : Internal oscillator is used for system clock
*/


    TRISC = TRISC | 0b00011000;              // Register Port C (0: Output, 1: Input)
        /*            +--------------------- Bit 7 Port RC7: here is output     (Pin 26)
                       +-------------------- Bit 6 Port RC6: here is output     (Pin 25)
                        +------------------- Bit 5 Port RC5: here is output     (Pin 24)
                         +------------------ Bit 4 Port RC4: here is input (SDA)(Pin 23)
                          +----------------- Bit 3 Port RC3: here is input (SCL)(Pin 18)
                           +---------------- Bit 2 Port RC2: here is output     (Pin 17)
                            +--------------- Bit 1 Port RC1: here is output     (Pin 16)
                             +-------------- Bit 0 Port RC0: here is output     (Pin 15)
*/


    // I2C-Hardware-Modul as Slave configuration
    SSPCON = 0b00100110;             // Sync Serial Port Control Register 1
/*             +-------------------- Bit 7 (WCOL): Write Collision Detect bit
                |||||||                   Master mode:
                |||||||                   0 : No collision
                |||||||                   1 : A write to SSPBUF was attempt while the I2C
                |||||||                       conditions were not valis
                |||||||                   Slave mode:
                |||||||                   0 : No collision
                |||||||                   1 : SSPBUF register is written while still
                |||||||                       transmitting the previous word (must be
                |||||||                       cleared in software)
                +------------------- Bit 6 (SSPOV): Receive Overflow Indicator bit
                 ||||||                   0 : No overflow
                 ||||||                   1 : A new byte is received while SSPBUF holds
                 ||||||                       previous data (weiter siehe Datenblatt!!)
                 ||||||                       (must be cleared in software)
                 +------------------ Bit 5 (SSPEN): Synchronous Serial Port Enable bit
                  |||||                   0 : Disables serial port and configures these
                  |||||                       Pins as I/O port pins
                  |||||              ->   1 : Enables the serial port an configures
                  |||||                       SCK, SDO, SDI, SS / SDA, SCL pins as the
                  |||||                       source of the serial port pins
                  +----------------- Bit 4 (CKP): Clock Polarity Select bit
                   ||||                   In SPI mode:
                   ||||                   0 : Idle state for clock is a low level
                   ||||                   1 : Idle state for clock is a high level
                   ||||                   In I2C Slave mode: SCK release control
                   ||||              ->   0 : Holds clock low (clock stretch). (Used to
                   ||||                       ensure data setup time.)
                   ||||                   1 : Enable clock
                   ||||                   In I2C Master mode: Unused in this mode
                   ++++------------- Bit 3-0 (SSPM3:SSPM0): SSP Mode Select bits
                                        0000 : SPI Master mode, clock = Fosc/4
                                        0001 : SPI Master mode, clock = Fosc/16
                                        0010 : SPI Master mode, clock = Fosc/64
                                        0011 : SPI Master mode, clock = TMR2 output/2
                                        0100 : SPI Slave mode, clock = SCK pin, nSS pin
                                               control enabled
                                        0101 : SPI Slave mode, clock = SCK pin, nSS pin
                                               control disabled. nSS can be used as I/O
                                     -> 0110 : I2C Slave mode, 7-bit address
                                        0111 : I2C Slave mode, 10-bit address
                                        1000 : I2C Master mode,
                                               clock = Fosc/(4*(SSPADD+1))
                                        1001 : Reserved
                                        1010 : Reserved
                                        1011 : I2C Firmware Controlled Master mode
                                               (slave idle)
                                        1100 : Reserved
                                        1101 : Reserved
                                        1110 : I2C Firmware Controlled Master mode,
                                               7-bit address with START and STOP bit
                                               interrupts enabled
                                        1111 : I2C Firmware Controlled Master mode,
                                               10-bit address with START and STOP bit
                                               interrupts enabled
*/

    SSPCON2 = 0b00110110;            // Sync Serial Port Control Register 2
/*              +------------------- Bit 7 (GCEN): General Call Enable bit (In I2C Slave
                 |||||||                  mode only)
                 |||||||             ->   0 : General Call address disabled
                 |||||||                  1 : Enable interrupt when a general call
                 |||||||                      address (0000h) ist received in the SSPSR
                 +------------------ Bit 6 (ACKSTAT) : Acknowledge Status bit (In I2C
                  ||||||                  Master mode only)
                  ||||||                  0 : Acknowledge was received from slave
                  ||||||                  1 : Acknowledge was not received from slave
                  +----------------- Bit 5 (ACKDT): Acknowledge Data bit (In I2C Master
                   |||||                  mode only):
                   |||||                  Value that will be transmitted when the user
                   |||||                  iniziates an Acknowledge sequence at the end
                   |||||                  of a receive
                   |||||                  0 : Acknowledge
                   |||||                  1 : Not Acknowledge
                   +---------------- Bit 4 (ACKEN): Acknowledge Sequence Enable bit
                    ||||                  (In I2C Master mode only):
                    ||||                  0 : Acknowledge sequence idle
                    ||||                  1 : Initiate Acknowledge sequence on SDA and
                    ||||                      SCL pins and transmit ACKDT data bit.
                    ||||                      Automatically cleared by hardware
                    +--------------- Bit 3 (RCEN): Receive Enable bit (In I2C Master
                     |||                  mode only)
                     |||                  0 : Receive idle
                     |||                  1 : Enables Receive mode for I2C
                     +-------------- Bit 2 (PEN): STOP Condition Enable bit (In I2C
                      ||                  Master mode only)
                      ||                  0 : STOP condition idle
                      ||                  1 : Initiate STOP condition on SDA and SCL pins
                      ||                      Automatically cleared by hardware
                      +------------- Bit 1 (RSEN): Repeated START Condition Enable bit
                       |                  (In I2C Master mode only)
                       |                  0 : Repeated START condition idle
                       |                  1 : Initiate Repeated START condition on SDA
                       |                      and SCL pins
                       |                      Automatically cleared by hardware
                       +------------ Bit 0 (SEN): START Condition Enable bit (In I2C
                                          Master mode only)
                                          0 : START condition idle
                                          1 : Initiate START cond. on SDA and SCL pins
                                             Automatically cleared by hardware
*/

     SSPSTAT = 0b10000000;
     // Set I2C Slave Adresse from _I2C_CHIP_ADDRESS constant
     cI2CAddress = _I2C_CHIP_ADDRESS;
     cI2CAddress = cI2CAddress << 1;       //Move address to left one bit
     cI2CAddress.F0 = 0;                   // This is a write from the master perspective
     SSPADD = cI2CAddress;


     // Interrupt flag
     PIE1 = PIE1 | 0b00001000;               // Peripheral Interrupt Enable Register 1 (Bank 1)
        /*           +---------------------- Bit 7  Reserve
                      +--------------------- Bit 6 (ADIE): A/D Converter Interrupt Enable bit
                       ||||||                ->   0 : Disables the ADC interrupt
                       ||||||                     1 : Enables the ADC interrupt
                       +-------------------- Bit 5 (RCIE): EUSART Receive Interrupt Enable bit
                        |||||                ->   0 : Disables the EUSART Receive Interrupt
                        |||||                     1 : Enables the EUSART Receive Interrupt
                        +------------------- Bit 4 (TXIE): EUSART Transmit Interrupt Enable bit
                         ||||                ->   0 : Disables the Transmit Interrupt Enable bit
                         ||||                     1 : Enables the Transmit Interrupt Enable bit
                         +------------------ Bit 3 (SSPIE): Master Synchronous Serial Port (MSSP)
                          |||                               Interrupt Enable bit
                          |||                     0 : Disables the MSSP interrupt
                          |||                ->   1 : Enables the MSSP interrupt
                          +----------------- Bit 2 (CCP1IF): CCP1 Interrupt Enable bit
                           ||                ->   0 : Disables the CCP1 interrupt
                           ||                     1 : Enables the CCP1 interrupt
                           +---------------- Bit 1 (TMR2IE): Timer 2 to PR2 Match Interrupt Enable bit
                            |                ->   0 : Disables the Timer 2 to PR2 Match Interrupt
                            |                     1 : Enables the Timer 2 to PR2 Match Interrupt
                            +--------------- Bit 0 (TMR1IE): Timer 1 Overflow Interrupt Enable bit
                                             ->   0 : Disables Timer 1 Overflow Interrupt
                                                  1 : Enables Timer 1 Overflow Interrupt
*/

     INTCON = INTCON | 0b11000000;        // Interrupt-Control-Register (Bank 1)
          /*             +-------------------- Bit 7 (GIE): Global Interrupt Enable bit
                          |||||||                   0 : Disables all interrupts
                          |||||||              ->   1 : Enables all unmasked interrupts
                          +------------------- Bit 6 (PEIE): Peripheral Interrupt Enable bit
                           ||||||                   0 : Disables all peripheral interrupts
                           ||||||              ->   1 : Enables all unmasked peripheral interrupts
                           +------------------ Bit 5 (T0IE): Timer 0 Overflow Interrupt Enable Bit
                            |||||              ->   0 : Disabled
                            |||||                   1 : Enabled
                            +----------------- Bit 4 (INTE): RB0/INT External Interrupt Enable Bit
                             ||||              ->   0 : Disabled
                             ||||                   1 : Enabled
                             +---------------- Bit 3 (RPIE): RB Port Change Interrupt Enable Bit
                              |||              ->   0 : Disabled
                              |||                   1 : Enabled
                              +--------------- Bit 2 (T0IF): Timer 0 Overflow Interrupt Flag bit
                               ||              ->   0 : TMR0 register did not overflow
                               ||                   1 : TMR0 register has overflowed
                               +-------------- Bit 1 (INTF): RB0/INT External Interrupt Flag bit
                                |              ->   0 : RB0/INT external interrupt did not occur
                                |                   1 : RB0/INT external interrupt occurred
                                +------------- Bit 0 (RBIF): RB Port Change Interrupt Flag bit
                                               ->   0 : None of RB7:RB4 pins have changed state
                                                    1 : One of the RB7:RB4 pins changed state
*/
}